from typing import Optional

from momento_wire_types.controlclient_pb2 import _CreateCacheRequest
from momento_wire_types.controlclient_pb2 import _DeleteCacheRequest
from momento_wire_types.controlclient_pb2 import _ListCachesRequest

from .._utilities._data_validation import _validate_cache_name
from ..cache_operation_responses import CreateCacheResponse
from ..cache_operation_responses import DeleteCacheResponse
from ..cache_operation_responses import ListCachesResponse

from .. import _cache_service_errors_converter
from .. import _momento_logger
from . import _scs_grpc_manager

_DEADLINE_SECONDS = 60.0  # 1 minute


class _ScsControlClient:
    """Momento Internal."""

    def __init__(self, auth_token: str, endpoint: str):
        self._grpc_manager = _scs_grpc_manager._ControlGrpcManager(auth_token, endpoint)

    async def create_cache(self, cache_name: str) -> CreateCacheResponse:
        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Creating cache with name: {cache_name}")
            request = _CreateCacheRequest()
            request.cache_name = cache_name
            return CreateCacheResponse(
                await self._grpc_manager.async_stub().CreateCache(
                    request, timeout=_DEADLINE_SECONDS
                )
            )
        except Exception as e:
            _momento_logger.debug(
                f"Failed to create cache: {cache_name} with exception:{e}"
            )
            raise _cache_service_errors_converter.convert(e) from None

    async def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Deleting cache with name: {cache_name}")
            request = _DeleteCacheRequest()
            request.cache_name = cache_name
            return DeleteCacheResponse(
                await self._grpc_manager.async_stub().DeleteCache(
                    request, timeout=_DEADLINE_SECONDS
                )
            )
        except Exception as e:
            _momento_logger.debug(
                f"Failed to delete cache: {cache_name} with exception:{e}"
            )
            raise _cache_service_errors_converter.convert(e) from None

    async def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        try:
            list_caches_request = _ListCachesRequest()
            list_caches_request.next_token = (
                next_token if next_token is not None else ""
            )
            return ListCachesResponse(
                await self._grpc_manager.async_stub().ListCaches(
                    list_caches_request, timeout=_DEADLINE_SECONDS
                )
            )
        except Exception as e:
            raise _cache_service_errors_converter.convert(e)

    async def close(self) -> None:
        await self._grpc_manager.close()



from typing import Union, Optional

from momento_wire_types.cacheclient_pb2 import _GetRequest
from momento_wire_types.cacheclient_pb2 import _SetRequest

from .. import cache_operation_responses as cache_sdk_resp
from .. import _cache_service_errors_converter
from .. import _momento_logger
from . import _scs_grpc_manager
from .._utilities._data_validation import (
    _as_bytes,
    _validate_ttl,
    _make_metadata,
    _validate_cache_name,
)


_DEFAULT_DEADLINE_SECONDS = 5.0  # 5 seconds


class _ScsDataClient:
    """Internal"""

    def __init__(
        self,
        auth_token: str,
        endpoint: str,
        default_ttl_seconds: int,
        operation_timeout_ms: Optional[int],
    ):
        self._default_deadline_seconds = (
            _DEFAULT_DEADLINE_SECONDS
            if not operation_timeout_ms
            else operation_timeout_ms / 1000.0
        )
        self._grpc_manager = _scs_grpc_manager._DataGrpcManager(auth_token, endpoint)
        _validate_ttl(default_ttl_seconds)
        self._default_ttlSeconds = default_ttl_seconds

    async def set(
        self,
        cache_name: str,
        key: str,
        value: Union[str, bytes],
        ttl_seconds: Optional[int],
    ) -> cache_sdk_resp.CacheSetResponse:
        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Issuing a set request with key {key}")
            item_ttl_seconds = (
                self._default_ttlSeconds if ttl_seconds is None else ttl_seconds
            )
            _validate_ttl(item_ttl_seconds)
            set_request = _SetRequest()
            set_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            set_request.cache_body = _as_bytes(value, "Unsupported type for value: ")
            set_request.ttl_milliseconds = item_ttl_seconds * 1000
            response = await self._grpc_manager.async_stub().Set(
                set_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            _momento_logger.debug(f"Set succeeded for key: {key}")
            return cache_sdk_resp.CacheSetResponse(response, set_request.cache_body)
        except Exception as e:
            _momento_logger.debug(f"Set failed for {key} with response: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def get(self, cache_name: str, key: str) -> cache_sdk_resp.CacheGetResponse:
        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Issuing a get request with key {key}")
            get_request = _GetRequest()
            get_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            response = await self._grpc_manager.async_stub().Get(
                get_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            _momento_logger.debug(f"Received a get response for {key}")
            return cache_sdk_resp.CacheGetResponse(response)
        except Exception as e:
            _momento_logger.debug(f"Get failed for {key} with response: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def close(self) -> None:
        await self._grpc_manager.close()



from typing import Any

import grpc
from grpc import ClientCallDetails


class _GenericClientInterceptor(grpc.UnaryUnaryClientInterceptor):
    def __init__(self, interceptor_function: Any):  # type: ignore[misc]
        self._fn = interceptor_function

    def intercept_unary_unary(  # type: ignore[misc]
        self,
        continuation: Any,
        client_call_details: ClientCallDetails,
        request: Any,
    ) -> Any:
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False
        )
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response


def create(intercept_call: Any) -> _GenericClientInterceptor:  # type: ignore[misc]
    return _GenericClientInterceptor(intercept_call)



from types import TracebackType
from typing import Optional, Union, Type

from ._scs_control_client import _ScsControlClient
from ._scs_data_client import _ScsDataClient
from .._utilities._data_validation import _validate_request_timeout

from .. import _momento_endpoint_resolver
from ..cache_operation_responses import (
    CreateCacheResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    CacheSetResponse,
    CacheGetResponse,
)


class SimpleCacheClient:
    """async Simple Cache client"""

    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        data_client_operation_timeout_ms: Optional[int],
    ):
        endpoints = _momento_endpoint_resolver.resolve(auth_token)
        self._control_client = _ScsControlClient(auth_token, endpoints.control_endpoint)
        self._data_client = _ScsDataClient(
            auth_token,
            endpoints.cache_endpoint,
            default_ttl_seconds,
            data_client_operation_timeout_ms,
        )

    async def __aenter__(self) -> "SimpleCacheClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self._control_client.close()
        await self._data_client.close()

    async def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a new cache in your Momento account.

        Args:
            cache_name: String used to create cache.

        Returns:
            CreateCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            ExistsError: If cache with the given name already exists.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.create_cache(cache_name)

    async def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all of the items within it.

        Args:
            cache_name: String cache name to delete.

        Returns:
            DeleteCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            NotFoundError: If an attempt is made to delete a MomentoCache that doesn't exits.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.delete_cache(cache_name)

    async def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        """Lists all caches.

        Args:
            next_token: Token to continue paginating through the list. It's used to handle large paginated lists.

        Returns:
            ListCachesResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
        """
        return await self._control_client.list_caches(next_token)

    async def set(
        self,
        cache_name: str,
        key: str,
        value: Union[str, bytes],
        ttl_seconds: Optional[int] = None,
    ) -> CacheSetResponse:
        """Stores an item in cache

        Args:
            cache_name: Name of the cache to store the item in.
            key (string or bytes): The key to be used to store item.
            value (string or bytes): The value to be stored.
            ttl_seconds (Optional): Time to live in cache in seconds. If not provided, then default TTL for the cache client instance is used.

        Returns:
            CacheSetResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to store the item.
        """
        return await self._data_client.set(cache_name, key, value, ttl_seconds)

    async def get(self, cache_name: str, key: str) -> CacheGetResponse:
        """Retrieve an item from the cache

        Args:
            cache_name: Name of the cache to get the item from
            key (string or bytes): The key to be used to retrieve the item.

        Returns:
            CacheGetResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        return await self._data_client.get(cache_name, key)


def init(
    auth_token: str,
    item_default_ttl_seconds: int,
    request_timeout_ms: Optional[int] = None,
) -> SimpleCacheClient:
    """Creates an async SimpleCacheClient

    Args:
        auth_token: Momento Token to authenticate the requests with Simple Cache Service
        item_default_ttl_seconds: A default Time To Live in seconds for cache objects created by this client. It is possible to override this setting when calling the set method.
        request_timeout_ms: An optional timeout in milliseconds to allow for Get and Set operations to complete. Defaults to 5 seconds. The request will be terminated if it takes longer than this value and will result in TimeoutError.
    Returns:
        SimpleCacheClient
    Raises:
        IllegalArgumentError: If method arguments fail validations.
    """
    _validate_request_timeout(request_timeout_ms)
    return SimpleCacheClient(auth_token, item_default_ttl_seconds, request_timeout_ms)



from typing import Callable, Union

import grpc
from grpc.aio import Metadata


class AddHeaderClientInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    def __init__(self, header_name: str, header_value: str):
        self._header_name = header_name
        self._header_value = header_value

    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> Union[grpc.aio._call.UnaryUnaryCall, grpc.aio._typing.ResponseType]:
        if client_call_details.metadata is None:
            client_call_details.metadata = Metadata()
        client_call_details.metadata.add(self._header_name, self._header_value)

        return await continuation(client_call_details, request)



from enum import Enum
from typing import Any, Optional, List

from momento_wire_types import cacheclient_pb2 as cache_client_types
from . import _cache_service_errors_converter as error_converter
from . import _momento_logger


class CacheGetStatus(Enum):
    HIT = 1
    MISS = 2


class CacheSetResponse:
    def __init__(self, grpc_set_response: Any, value: bytes):  # type: ignore[misc]
        """Initializes CacheSetResponse to handle gRPC set response.

        Args:
            grpc_set_response: Protobuf based response returned by Scs.
            value (string or bytes): The value to be used to store item in the cache

        Raises:
            InternalServerError: If server encountered an unknown error while trying to store the item.
        """
        self._value = value

    def value(self) -> str:
        """Decodes string value set in cache to a utf-8 string."""
        return self._value.decode("utf-8")

    def value_as_bytes(self) -> bytes:
        """Returns byte value set in cache."""
        return self._value


class CacheGetResponse:
    def __init__(self, grpc_get_response: Any):  # type: ignore[misc]
        """Initializes CacheGetResponse to handle gRPC get response.

        Args:
            grpc_get_response: Protobuf based response returned by Scs.

        Raises:
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        self._value: bytes = grpc_get_response.cache_body

        if grpc_get_response.result == cache_client_types.Hit:
            self._result = CacheGetStatus.HIT
        elif grpc_get_response.result == cache_client_types.Miss:
            self._result = CacheGetStatus.MISS
        else:
            _momento_logger.debug(
                f"Get received unsupported ECacheResult: {grpc_get_response.result}"
            )
            raise error_converter.convert_ecache_result(
                grpc_get_response.result, grpc_get_response.message, "GET"
            )

    def value(self) -> Optional[str]:
        """Returns value stored in cache as utf-8 string if there was Hit. Returns None otherwise."""
        if self._result == CacheGetStatus.HIT:
            return self._value.decode("utf-8")
        return None

    def value_as_bytes(self) -> Optional[bytes]:
        """Returns value stored in cache as bytes if there was Hit. Returns None otherwise."""
        if self._result == CacheGetStatus.HIT:
            return self._value
        return None

    def status(self) -> CacheGetStatus:
        """Returns get operation result such as HIT or MISS."""
        return self._result


class CreateCacheResponse:
    def __init__(self, grpc_create_cache_response: Any):  # type: ignore[misc]
        pass


class DeleteCacheResponse:
    def __init__(self, grpc_delete_cache_response: Any):  # type: ignore[misc]
        pass


class CacheInfo:
    def __init__(self, grpc_listed_cache: Any):  # type: ignore[misc]
        """Initializes CacheInfo to handle caches returned from list cache operation.

        Args:
            grpc_listed_cache: Protobuf based response returned by Scs.
        """
        self._name: str = grpc_listed_cache.cache_name

    def name(self) -> str:
        """Returns all cache's name."""
        return self._name


class ListCachesResponse:
    def __init__(self, grpc_list_cache_response: Any):  # type: ignore[misc]
        """Initializes ListCacheResponse to handle list cache response.

        Args:
            grpc_list_cache_response: Protobuf based response returned by Scs.
        """
        self._next_token: Optional[str] = (
            grpc_list_cache_response.next_token
            if grpc_list_cache_response.next_token != ""
            else None
        )
        self._caches = []
        for cache in grpc_list_cache_response.cache:
            self._caches.append(CacheInfo(cache))

    def next_token(self) -> Optional[str]:
        """Returns next token."""
        return self._next_token

    def caches(self) -> List[CacheInfo]:
        """Returns all caches."""
        return self._caches



from ._generic_client_interceptor import _GenericClientInterceptor

from . import _header_client_interceptor


def get_authorization_interceptor(auth_token: str) -> _GenericClientInterceptor:
    return _header_client_interceptor.header_adder_interceptor(
        "authorization", auth_token
    )



import asyncio
from typing import Any


# NOTES:
#
# 1. it would be really nice to put some more meaningful type hints here but I'm not familiar
#    enough with asyncio / coroutines to understand their types.
# 2. Kenny believes that we should be able to do this without a direct dependency on asyncio,
#    by writing a loop that calls `send` on the coroutine and then returns on StopIteration.
#    I was not able to get this working during my timeboxed window so left it like this for now.
def wait_for_coroutine(coroutine: Any) -> Any:  # type: ignore[misc]
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine)  # type: ignore[misc]



import setuptools
import os

version = os.getenv("MOMENTO_SDK_VERSION")

if(version is None):
    version = '0.0.0+localdev0'

# version is the only dynamic configuration
setuptools.setup(
    version=version,
    long_description="# Momento SDK\n Check out our SDK example [here](https://github.com/momentohq/client-sdk-examples/tree/main/python)!",
    long_description_content_type="text/markdown",
    project_urls={
        'Source': 'https://github.com/momentohq/client-sdk-python',
    },
)



class SdkError(Exception):
    """Base exception for all errors raised by Sdk"""

    def __init__(self, message: str):
        super().__init__(message)


class ClientSdkError(SdkError):
    """For all errors raised by the client.

    Indicates that the request failed on the SDK. The request either did not
    make it to the service or if it did the response from the service could
    not be parsed successfully.
    """

    def __init__(self, message: str):
        super().__init__(message)


class InvalidArgumentError(ClientSdkError):
    """Error raised when provided input values to the SDK are invalid

    Some examples - missing required parameters, incorrect parameter
    types, malformed input.
    """

    def __init__(self, message: str):
        super().__init__(message)


class MomentoServiceError(SdkError):
    """Errors raised when Momento Service returned an error code"""

    def __init__(self, message: str):
        super().__init__(message)


class NotFoundError(MomentoServiceError):
    """Error raised for operations performed on non-existent cache"""

    def __init__(self, message: str):
        super().__init__(message)


class AlreadyExistsError(MomentoServiceError):
    """Error raised when attempting to create a cache with same name"""

    def __init__(self, message: str):
        super().__init__(message)


class BadRequestError(MomentoServiceError):
    """Error raised when service validation fails for provided values"""

    def __init__(self, message: str):
        super().__init__(message)


class PermissionError(MomentoServiceError):
    """Error for insufficient permissions to perform an operation with Cache Service."""

    def __init__(self, message: str):
        super().__init__(message)


class AuthenticationError(MomentoServiceError):
    """Error when authentication with Cache Service fails"""

    def __init__(self, message: str):
        super().__init__(message)


class CancelledError(MomentoServiceError):
    """Error when an operation with Cache Service was cancelled"""

    def __init__(self, message: str):
        super().__init__(message)


class TimeoutError(MomentoServiceError):
    """Error when an operation did not complete in time"""

    def __init__(self, message: str):
        super().__init__(message)


class LimitExceededError(MomentoServiceError):
    """Error when calls are throttled due to request limit rate"""

    def __init__(self, message: str):
        super().__init__(message)


class InternalServerError(MomentoServiceError):
    """Operation failed on the server with an unknown error"""

    def __init__(self, message: str):
        super().__init__(message)



import grpc

import momento_wire_types.cacheclient_pb2_grpc as cache_client
import momento_wire_types.controlclient_pb2_grpc as control_client

from ._add_header_client_interceptor import AddHeaderClientInterceptor


class _ControlGrpcManager:
    """Momento Internal."""

    def __init__(self, auth_token: str, endpoint: str):
        self._secure_channel = grpc.aio.secure_channel(
            target=endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=[AddHeaderClientInterceptor("authorization", auth_token)],
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> control_client.ScsControlStub:
        return control_client.ScsControlStub(self._secure_channel)


class _DataGrpcManager:
    """Momento Internal."""

    def __init__(self, auth_token: str, endpoint: str):
        self._secure_channel = grpc.aio.secure_channel(
            target=endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=[AddHeaderClientInterceptor("authorization", auth_token)],
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> cache_client.ScsStub:
        return cache_client.ScsStub(self._secure_channel)



import collections
from typing import Any

import grpc
from ._generic_client_interceptor import _GenericClientInterceptor

from . import _generic_client_interceptor


class _ClientCallDetails(
    collections.namedtuple(
        "_ClientCallDetails", ("method", "timeout", "metadata", "credentials")
    ),
    grpc.ClientCallDetails,
):
    pass


def header_adder_interceptor(header: str, value: str) -> _GenericClientInterceptor:
    def intercept_call(  # type: ignore[misc]
        client_call_details: Any,
        request_iterator: Any,
        request_streaming: Any,
        response_streaming: Any,
    ) -> Any:
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append(
            (
                header,
                value,
            )
        )
        client_call_details = _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
        )
        return client_call_details, request_iterator, None

    return _generic_client_interceptor.create(intercept_call)
